@@ -1,7 +1,11 @@
 import os
 import json
+import functools
+import contextlib
 import traceback
 import requests
+import urllib
+from . import RequestError
 
 
 HOST_KEY = 'VIRTUAL_HOST'
@@ -10,11 +14,6 @@ DEFAULT_HOST = 'localhost'
 DEFAULT_PORT = 8000
 
 
-class RequestError(RuntimeError):
-    pass
-
-
-
 def get_well_known(url, realm=None, secure=None):
     '''Get the well known for an oauth2 server.
 
@@ -62,7 +61,7 @@ def get_well_known(url, realm=None, secure=None):
     '''
     if not url.startswith('https://'):
         parts = url.split('@', 1)
-        url = as_url('{}/auth/realms/{}/.well-known/openid-configuration'.format(
+        url = asurl('{}/auth/realms/{}/.well-known/openid-configuration'.format(
             parts[-1], realm or (parts[0] if len(parts) > 1 else 'master')), secure=secure)
 
     resp = requests.get(url).json()
@@ -109,19 +108,19 @@ def with_keycloak_secrets_file(
     })
 
 
-def with_keycloak_secrets_file_from_environment(prefix, url=None, realm=None, fname=None):
+def with_keycloak_secrets_file_from_environment(env=None, url=None, realm=None, fname=None):
+    env = env or 'APP'
+    if isinstance(env, str):
+        env = Env(env)
     return with_keycloak_secrets_file(
-        as_url(url or envv('AUTH_HOST', prefix)),
-        envv('CLIENT_ID', prefix),
-        envv('CLIENT_SECRET', prefix),
-        realm=realm or envv('AUTH_REALM', prefix, 'master'),
-        redirect_uris=get_redirect_uris(envv('REDIRECT_URIS', prefix)),
+        asurl(url or env('AUTH_HOST')), env('CLIENT_ID'), env('CLIENT_SECRET'),
+        realm=realm or env('AUTH_REALM', 'master'),
+        redirect_uris=get_redirect_uris(env('REDIRECT_URIS')),
         fname=fname,
     )
 
 
 def _write_secrets_file(fname, cfg):
-    print(fname, cfg)
     if not fname:
         return cfg
     if fname is True:
@@ -136,28 +135,155 @@ def _write_secrets_file(fname, cfg):
 
 
 
-def kmerge(*xs, sep='_'):
-    return sep.join(str(x) for x in xs if x)
+# def kmerge(*xs, sep='_'):
+#     return sep.join(str(x) for x in xs if x)
 
-def envv(k, prefix=None, default=None):
-    return os.getenv(kmerge(k, prefix)) or default
+# def envv(k, prefix=None, default=None):
+#     return os.getenv(kmerge(k, prefix)) or default
+
+def aslist(x):
+    return x if isinstance(x, (list, tuple)) else [x] if x else []
 
 
-
-def as_url(url, *paths, secure=None):
+def asurl(url, *paths, secure=None, **args):
     if url:
-        if not url.startswith('http://') or url.startswith('https://'):
+        if not (url.startswith('http://') or url.startswith('https://')):
             if secure is None:
                 secure = url != 'localhost'
             url = 'http{}://{}'.format(bool(secure)*'s', url)
-        return os.path.join(url, *(p.lstrip('/') for p in paths))
+        url = os.path.join(url, *(p.lstrip('/') for p in paths))
+        # add args, and make sure they go before the hashstring
+        args = {k: v for k, v in args.items() if v is not None}
+        if args:
+            url, hsh = url.split('#', 1) if '#' in url else (url, '')
+            url += ('&' if '?' in url else '?') + urllib.parse.urlencode(args)
+            if hsh:
+                url += '#' + hsh
+        return url
+as_url = asurl  # backwards compat
 
 def get_redirect_uris(uris=None):
-    uris = as_url(uris or os.getenv(HOST_KEY) or '{}:{}/*'.format(DEFAULT_HOST, os.getenv(PORT_KEY, str(DEFAULT_PORT))))
+    uris = asurl(uris or os.getenv(HOST_KEY) or '{}:{}/*'.format(DEFAULT_HOST, os.getenv(PORT_KEY, str(DEFAULT_PORT))))
     return uris if isinstance(uris, (list, tuple)) else [uris] if uris else uris
 
 
-def traceback_html(e):
-    return '''{err}<pre><h3>Traceback (most recent call last):</h3><div>{tb}</div><h3>{typename}: {err}</h3></pre>'''.strip().format(
-        err=e, tb='<hr/>'.join(traceback.format_tb(e.__traceback__)).strip('\n'),
-        typename=type(e).__name__)
+# def traceback_html(e):
+#     return '''{err}<pre><h3>Traceback (most recent call last):</h3><div>{tb}</div><h3>{typename}: {err}</h3></pre>'''.strip().format(
+#         err=e, tb='<hr/>'.join(traceback.format_tb(e.__traceback__)).strip('\n'),
+#         typename=type(e).__name__)
+
+
+
+
+class Env:
+    def __init__(self, prefix=None, upper=True, **kw):
+        self.prefix = prefix or ''
+        self.upper = upper
+        self.vars = kw
+        if self.upper:
+            self.prefix = self.prefix.upper()
+
+    def __str__(self):
+        return '<env {}>'.format(''.join([
+            '\n  {}={}'.format(k, self(k)) for k in self.vars
+        ]))
+
+    def __contains__(self, key):
+        return self.key(key) in os.environ
+
+    def __getattr__(self, key):
+        return self.get(key)
+
+    def __call__(self, *keys, **kw):
+        return (
+            (self.get(keys[0], **kw) if len(keys) == 1 else [self.get(k, **kw) for k in keys])
+            if keys else {k: self.get(v) for k, v in kw.items()}
+        )
+
+    def get(self, key, default=None, cast=None):
+        y = os.environ.get(self.key(key))
+        if y is None:
+            return default
+        if callable(cast):
+            return y
+        if y in ('1', '0'):
+            y = int(y)
+        if y.lower() in ('y', 'n'):
+            y = y.lower() == 'y'
+        return y
+
+    def key(self, x):
+        k = (self.prefix or '') + self.vars.get(x, x)
+        return k.upper() if self.upper else k
+
+    def all(self):
+        return {k: v for k, v in os.environ.items() if k.startswith(self.prefix)}
+
+
+
+
+
+class Colors(dict):
+    '''Color text. e.g.
+    >>> print(color('hi', 'red') + color.blue('hello') + color['green']('sup'))
+    '''
+    def __call__(self, x, name=None):
+        if not name:
+            return str(x)
+        return '\033[{}m{}\033[0m'.format(super().__getitem__(name.lower()), x) if name else x
+    def __getitem__(self, k):
+        if k not in self:
+            raise KeyError(k)
+        return functools.partial(self.__call__, name=k)
+    def __getattr__(self, k):
+        if k not in self:
+            raise AttributeError(k)
+        return functools.partial(self.__call__, name=k)
+
+color = Colors(
+    black='0;30',
+    red='0;31',
+    green='0;32',
+    orange='0;33',
+    blue='0;34',
+    purple='0;35',
+    cyan='0;36',
+    lightgray='0;37',
+    darkgray='1;30',
+    lightred='1;31',
+    lightgreen='1;32',
+    yellow='1;33',
+    lightblue='1;34',
+    lightpurple='1;35',
+    lightcyan='1;36',
+    white='1;37',
+)
+color_ = color
+
+
+def ask(question, color=None, secret=False):
+    prompt = input
+    if secret:
+        import getpass
+        prompt = getpass.getpass
+    return prompt(':: {} '.format(color_(question, color)))
+
+
+@contextlib.contextmanager
+def saveddict(fname):
+    import base64
+    try:
+        data = {}
+        fname = fname and os.path.expanduser(fname)
+        if fname and os.path.isfile(fname):
+            try:
+                with open(fname, 'rb') as f:
+                    data = json.loads(base64.b64decode(f.read()).decode('utf-8'))
+            except json.decoder.JSONDecodeError:
+                pass
+        yield data
+    finally:
+        if fname:
+            os.makedirs(os.path.dirname(fname) or '.', exist_ok=True)
+            with open(fname, 'wb') as f:
+                f.write(base64.b64encode(json.dumps(data).encode('utf-8')))
